with Globals, Strings, Ints, Flts, Reps, Reps.Balls;
use Globals, Strings, Ints, Flts, Reps, Reps.Balls;
with Matrices, Vectors, ScalVectors, Hopf.Fix;

pragma Elaborate_All(Globals,Strings,Ints,Flts,Reps.Balls,Matrices,Vectors,ScalVectors,Hopf.Fix);

procedure Check is

  NT: constant Positive := GetArg;                -- number of concurrent tasks
  Rho:  constant Radius := Flt(32);               -- domain of analyticity in time
  Beth: constant Radius := Half**12;              -- normalization constant
  UGc:  constant Radius := Flt(170339)/Flt(2048); -- center of gamma used for the stationary solution
  UGw:  constant Radius := Half**10;              -- radius of gamma used for the stationary solution

  NN:             Natural;   -- main frequency
  DegF:           Positive;  -- size of Fourier2
  Mat_PackSize:   Positive;  -- pack size for matrix used in Contr
  Contr_PackSize: Positive;  -- used in proof

  procedure Set(N:in  Natural) is
  begin
    NN := N;
    Show1("NN=",NN);
    if NN=0 then
      DegF           := 86;
      Mat_PackSize   := 64;
      Contr_PackSize := 86;
    elsif NN=1 then
      DegF           := 86;
      Mat_PackSize   := 64;
      Contr_PackSize := 86;
    elsif NN=2 then
      DegF           := 72;
      Mat_PackSize   := 54;
      Contr_PackSize := 70;
    elsif NN=3 then
      DegF           := 66;
      Mat_PackSize   := 54;
      Contr_PackSize := 64;
    elsif NN=4 then
      DegF           := 60;
      Mat_PackSize   := 52;
      Contr_PackSize := 54;
    elsif (NN mod 2) = 0 then
      DegF           := 54;
      Mat_PackSize   := 52;
      Contr_PackSize := 52;
    else
      DegF           := 56;
      Mat_PackSize   := 54;
      Contr_PackSize := 54;
    end if;
  end Set;

  subtype NumScalar is Rep;
  subtype Scalar is Ball;

  procedure CheckNN is

    NUmmaGumma: constant NumScalar := Scal(UGc);
    UmmaGumma: constant Scalar := Scal(UGc)+BallAt0(UGw);  --- [83.1724,83.1743]

    package NSV is new Vectors (Component => NumScalar);
    package NSM is new Matrices (Scalar => NumScalar);
    package NSSV is new ScalVectors (Scalar => NumScalar, SM => NSM, SV => NSV);
    package NHopf is new Hopf(NN => NN, DegF => DegF, Rho => Rho, Scalar => NumScalar);
    package NFix is new NHopf.Fix(SSV => NSSV, Mat_PackSize => Mat_PackSize, Beth => Beth,
                                  SV => NSV, SM => NSM, UmmaGumma => NUmmaGumma);

    package SV is new Vectors (Component => Scalar);
    package SM is new Matrices (Scalar => Scalar);
    package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);
    package Hopph is new Hopf(NN => NN, DegF => DegF, Rho => Rho, Scalar => Scalar);
    package Fix is new Hopph.Fix(SSV => SSV, Mat_PackSize => Mat_PackSize, Beth => Beth,
                                 SV => SV, SM => SM, UmmaGumma => UmmaGumma);

    NF: NHopf.Fooreeay.Fourier3;
    F:  Hopph.Fooreeay.Fourier3;
    NM:   NFix.QC_Matrix;
    M,M1: Fix.QC_Matrix;

  begin

    if (Hopph.Fou.NNEven and NN<293) or (Hopph.Fou.NNOdd and NN<304) then
      NHopf.Fooreeay.Read(Input_Dir & "BP.approx",NF);
      NFix.ContrMatrix(NF,NM);
      for J in M'Range(1) loop
        for K in M'Range(2) loop
          Assign(Approx(NM(J,K)),M(J,K));
          Copy(M(J,K),M1(J,K));
        end loop;
      end loop;
      SM.AddConst(1,M1);
      SSV.Invert(M1);
      Show0("Matrix ok");
    else
      SM.SetZero(M);
    end if;
    Hopph.Fooreeay.Read(Input_Dir & "BP.approx",F);
    Fix.ContrFix(M,Contr_PackSize,F);
  end CheckNN;

begin

  NTasks := NT;
  Stack_Size := 2**30;
  Verbosity := 2;

  for N in 0 .. 305 loop
    if N<295 or (N mod 2)=1 then
      Set(N);
      CheckNN;
    end if;
  end loop;

end Check;
