with Strings, Ints, Globals;
use Strings, Ints, Globals;

pragma Elaborate_All (Strings,Ints,Globals);

package body Hopf.Fix is

  use Lin,Lin_Contr;

  procedure AB(F: in Fourier3; A,B: in out Scalar) is
  begin
    pragma Warnings (Off);
    Coeff( 1,1,1,F,A);
    Coeff(-1,1,1,F,B);
    pragma Warnings (On);
  end AB;

  procedure Alpha_Gamma(P: in Fourier3; Alpha,Gamma: in out Scalar) is
    A,B: Scalar;
  begin
    if NN=0 then
      SetZero(Alpha);
      Copy(UmmaGumma,Gamma);
    else
      AB(P,A,B);
      Quot(STwo,B,Gamma);
      Prod(A,Gamma,Alpha);     ---  2*A/B
      Neg(Gamma);
      Mult(SqrtTwoBeth,Gamma); --- -2*SqrtTwoBeth/B
    end if;
  end Alpha_Gamma;

  procedure Normalize(A,B: in Flt; F: in out Fourier3) is
  begin
    pragma Warnings (Off);
    Assign(A,F.C(1,1)( 1));
    Assign(B,F.C(1,1)(-1));
    pragma Warnings (On);
  end Normalize;

  procedure Prod_Parallel(F1,F2: in Fourier3; F3: in out Fourier3) is
    Saved_FProd_Parallel: constant Boolean := FProd_Parallel;
  begin
    FProd_Parallel := True;
    Prod(F1,F2,F3);
    FProd_Parallel := Saved_FProd_Parallel;
  end Prod_Parallel;

  procedure GMap(F: in Fourier3; GF: in out Fourier3) is
    Alpha,Gamma: Scalar;
  begin
    Prod_Parallel(F,F,GF);
    Alpha_Gamma(GF,Alpha,Gamma);
    InvLinear(Alpha,GF);
    Mult(NegHalf,Gamma);
    Mult(Gamma,GF);
    Add(NegInvLapF,GF);
    if NN>0 then Normalize(Zero,Beth,GF); end if;
  end GMap;

  function PostFix return String is
  begin
    if STrunc then return ".approx"; else return ""; end if;
  end PostFix;

  procedure BeforeDG(F: in Fourier3; L0,L00: in out Fourier3; Alpha,Gamma: in out Scalar) is
    R: Scalar;
  begin
    if NN=1 then
      Prod_Parallel(F,F,L0);
      Alpha_Gamma(L0,Alpha,Gamma);
      Write(SZero,Alpha,Gamma);
      InvLinear(Alpha,L0);
      DtInvLinear(Alpha,L0,L00);
      Write(Output_Dir & "L0" & PostFix,L0);
      Write(Output_Dir & "L00" & PostFix,L00);
    elsif NN=2 then
      Read(Input_Dir & "L0" & PostFix,L0);
      Read(Input_Dir & "L00" & PostFix,L00);
    else
      SetZero(L0);
      SetZero(L00);
    end if;
    if NN=0 then
      SetZero(Alpha);
      Copy(UmmaGumma,Gamma);
    elsif NN>1 then
      Read(R,Alpha,Gamma);
    end if;
  end BeforeDG;

  procedure DAlpha_Gamma(DP: in Fourier3; Alpha,Gamma: in Scalar; DAlpha,DGamma: in out Scalar) is
    DA,DB,AA,BB,Tmp: Scalar;
  begin
    AB(DP,DA,DB);
    Quot(Gamma,SqrtTwoBeth,Tmp);
    Prod(DA,Tmp,AA);             -- AA = DA*Gamma/SqrtTwoBeth
    Prod(DB,Tmp,BB);             -- BB = DB*Gamma/SqrtTwoBeth
    Prod(Gamma,BB,DGamma);
    Mult(SHalf,DGamma);          -- DGamma = BB*Gamma/2
    Prod(Alpha,BB,DAlpha);
    Mult(SHalf,DAlpha);
    Sub(AA,DAlpha);              -- DAlpha = BB*Alpha/2 - AA
  end DAlpha_Gamma;

  procedure DGMap(F,H,L0,L00: in Fourier3; Alpha,Gamma: in Scalar; DGH: in out Fourier3) is
    Dg,Da: Scalar;
    L1: Fourier3;
  begin
    Prod(F,H,L1);
    InvLinear(Alpha,L1);
    Prod(Gamma,L1,DGH);  --- neg Dpsi term
    if NN=1 or else NN=2 then
      DAlpha_Gamma(L1,Alpha,Gamma,Da,Dg);
      Mult(Gamma,Da);
      Mult(SHalf,Da);      --- Gamma*DAlpha/2
      Mult(SHalf,Dg);      --- DGamma/2
      AddProd(Dg,L0,DGH);  --- add DGamma term
    end if;
    Neg(DGH);
    if NN=1 or else NN=2 then
      AddProd(Da,L00,DGH); --- add DAlpha term
    end if;
    if NN>0 then Normalize(Zero,Zero,DGH); end if;
  end DGMap;

  procedure DContrNorm(M: in SMatrix; F: in Fourier3; PackSize: in Positive; R: out Radius) is
    Q: constant F3Modes := Make(PackSize,True,False);  --- Err modes, but Extract is not used
    L0,L00: Fourier3;
    A,B,R1: Scalar;
    procedure DG(H: in Fourier3; DGH: in out Fourier3) is begin DGMap(F,H,L0,L00,A,B,DGH); end DG;
    procedure DContract is new DContr (DMap => DG);
    procedure DC(H: in Fourier3; DCH: in out Fourier3) is begin DContract(QC,M,H,DCH); end DC;
    procedure DCNorm is new Op_Norm (LinOp => DC);
    procedure DGNorm is new Op_Norm (LinOp => DG);
  begin
    if (PackSize <= Mat_PackSize) and (NN<2) then raise Not_Implemented; end if;
    if Verbosity>1 then
      Show1("DContrNorm: Q'Last = ",Q'Last);
    end if;
    if SM.IsZero(M) then
      SetZero(L0);
      SetZero(L00);
      Read(R1,A,B);
      DGNorm(Q,R);
    else
      BeforeDG(F,L0,L00,A,B);
      DCNorm(Q,R);
    end if;
  end DContrNorm;

  procedure ContrFix(M: in SMatrix; PackSize: in Positive; F: in out Fourier3) is
    KMax: constant Radius := Flt(7)/Flt(8);
    E,K: Radius;
    ANorm,R,Eps,Del,Alpha,Gamma: Scalar;
    F0: Fourier3;
  begin
    Copy(F,F0);
    Show0("ContrFix: ",False);
    if NN<2 then
      GMap(F0,F);
      Sub(F0,F);
      E := MaxNorm(F);
      Assign(E,Eps);
    else
      Read(Eps,Alpha,Gamma);
    end if;
    OpNorm(QC,M,ANorm);
    Add(SOne,ANorm);
    Show2("Eps,ANorm ",Sup(Eps),Sup(ANorm));
    ---
    Del := Eps/(SOne-Scal(KMax))+Scal(Tiny);
    R := ANorm*Del;
    Copy(F0,F);
    AddBall(MaxNorm(R),F);
    Show1("R ",Sup(R));
    DContrNorm(M,F,PackSize,K);
    Show1("ContrFix: K ",K);
    if NN<2 then
      if K>KMax then raise Sorry with "K too large"; end if;
    else
      if K+Tiny>One then raise Sorry with "K too large"; end if;
    end if;
    ---
    Del := Eps/(SOne-Scal(K))+Scal(Tiny);
    R := ANorm*Del;
    Show2("ContrFix: Del,R: ",Sup(Del),Sup(R));
    Copy(F0,F);
    AddBall(MaxNorm(R),F);
    if NN=1 then
      Normalize(Zero,Beth,F);
      Prod_Parallel(F,F,F0);
      Alpha_Gamma(F0,Alpha,Gamma);
      Show2("alpha,gamma",Alpha,Gamma);
      Write(R,Alpha,Gamma);
    end if;
  end ContrFix;

  procedure Read(R,A,G: in out Scalar) is
    F: File_Type;
  begin
    if Verbosity>1 then Put_Line("Reading R,alpha,gamma"); end if;
    Open(F,In_File,Input_Dir & "RAlphaGamma" & PostFix);
    Get(F,R);
    Get(F,A);
    Get(F,G);
    Close(F);
  end Read;

  procedure Write(R,A,G: in Scalar) is
    F: File_Type;
  begin
    if Verbosity>1 then Put_Line("Writing R,alpha,gamma"); end if;
    Create(F,Out_File,Output_Dir & "RAlphaGamma" & PostFix);
    Put(F,R);
    Put(F,A);
    Put(F,G);
    Close(F);
  end Write;

  --------------------------------------------
  --------------------------------------------

  procedure DerMatrix(F: in Fourier3; M: in out SMatrix) is
    --- for numeric mode only
    L0,L00: Fourier3;
    A,B: Scalar;
    procedure DG(H: in Fourier3; DGH: in out Fourier3) is begin DGMap(F,H,L0,L00,A,B,DGH); end DG;
    procedure Enclose is new Enclosure (LinOp => DG);
  begin
    if Not_STrunc then raise Sorry; end if;
    BeforeDG(F,L0,L00,A,B);
    Enclose(QC,M);
  end DerMatrix;

  procedure ContrMatrix(F: in Fourier3; M: in out SMatrix) is
  begin
    if Verbosity>1 then Show0("running ContrMatrix"); end if;
    DerMatrix(F,M);
    Contr_Matrix(M);
  end ContrMatrix;

  procedure ApproxFix(F: in out Fourier3; M: in out SMatrix; Eps: in Radius := Zero) is
    --- find approximate fixed point
    procedure Contract0 is new ContrTmp0 (Map => GMap);
    Iter: constant Integer := 1024;
    MaxCredit: constant Integer := 8;
    Credit: Integer := MaxCredit;
    Err,MinErr: Radius := 1.0E+9;
    --    ErrMat: constant Radius := 1.0e-3;
    G,Tmp1,Tmp2: Fourier3;
  begin
    if Not_STrunc then raise Sorry with "only for numeric mode"; end if;
    if Verbosity>1 then Show0("running ApproxFix"); end if;
    Copy(F,G);
    for I in 1 .. Iter loop
      --      if Err>ErrMat then ContrMatrix(G,M); end if;
      Contract0(QC,M,G,Tmp1,Tmp2,Err);
      Show1("ApproxFix: step " & Strng(I) & ", Err ",Err);
      if Err<MinErr then
        Copy(G,F);
        if (Err<Eps) or (Err=MinErr) then exit; end if;
        MinErr := Err;
        Credit := MaxCredit;
      else
        exit when Credit=0;
        Credit := Credit-1;
      end if;
    end loop;
    Show1("ApproxFix: Err ",Err);
    --      ContrMatrix(F,M);
  end ApproxFix;

  procedure IterContr(F: in out Fourier3; M: in out SMatrix) is
    --- improve approximate fixed point
    procedure Contract0 is new ContrTmp0 (Map => GMap);
    MaxCredit: constant Integer := 16;
    Credit: Integer := MaxCredit;
    Err,MinErr: Radius := One;
    G,Tmp1,Tmp2: Fourier3;
  begin
    if Not_STrunc then raise Sorry with "only for numeric mode"; end if;
    if Verbosity>1 then Show0("running IterContr"); end if;
    Copy(F,G);
    ContrMatrix(G,M);
    while Credit>0 loop
      Contract0(QC,M,G,Tmp1,Tmp2,Err);
      Show1("IterContr: Err ",Err);
      if Err<MinErr then
        Copy(G,F);
        MinErr := Err;
        Credit := MaxCredit;
      else
        Credit := Credit-1;
      end if;
    end loop;
    Show1("IterContr: Err ",MinErr);
  end IterContr;

end Hopf.Fix;
