with Vectors, Matrices, ScalVectors, Linear, Linear.Contr;

pragma Elaborate_All (ScalVectors,Linear.Contr);

generic

  Mat_PackSize: in Positive;
  Beth: in Radius;
  UmmaGumma: in Scalar := Scal(0);

  with package SM is new Matrices (Scalar => Scalar);
  with package SV is new Vectors (Component => Scalar);
  with package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);

package Hopf.Fix is

  subtype SMatrix is SM.Matrix;

  procedure Alpha_Gamma(P: in Fourier3; Alpha,Gamma: in out Scalar);
  procedure Normalize(A,B: in Flt; F: in out Fourier3);
  procedure GMap(F: in Fourier3; GF: in out Fourier3);
  procedure DAlpha_Gamma(DP: in Fourier3; Alpha,Gamma: in Scalar; DAlpha,DGamma: in out Scalar);
  procedure DGMap(F,H,L0,L00: in Fourier3; Alpha,Gamma: in Scalar; DGH: in out Fourier3);

  procedure DContrNorm(M: in SMatrix; F: in Fourier3; PackSize: in Positive; R: out Radius);
  procedure ContrFix(M: in SMatrix; PackSize: in Positive; F: in out Fourier3);

  package SV_Ops renames SSV.SV_Ops;
  package Lin is new Linear (Scalar => Scalar, SM => SM, SV => SV, Fun => Fourier3, Mode => F3Mode, Modes => F3Modes);
  package Lin_Contr is new Lin.Contr (SSV => SSV);

  --- numerical
  procedure DerMatrix(F: in Fourier3; M: in out SMatrix);
  procedure ContrMatrix(F: in Fourier3; M: in out SMatrix);
  procedure ApproxFix(F: in out Fourier3; M: in out SMatrix; Eps: in Radius := Zero);
  procedure IterContr(F: in out Fourier3; M: in out SMatrix);

  QC: F3Modes := Make(Mat_PackSize,True,True);   --- Simple and no Err modes
  subtype QC_Vector is SV.Vector(QC'Range);
  subtype QC_Matrix is SMatrix(QC'Range,QC'Range);

private

  SqrtTwo:     constant Scalar   := Sqrt(STwo);
  SqrtTwoBeth: constant Scalar   := Beth*SqrtTwo;
  NegInvLapF:  constant Fourier3 := Make_NegInvLapForce;

  procedure Prod_Parallel(F1,F2: in Fourier3; F3: in out Fourier3);
  function PostFix return String;
  procedure BeforeDG(F: in Fourier3; L0,L00: in out Fourier3; Alpha,Gamma: in out Scalar);
  procedure AB(F: in Fourier3; A,B: in out Scalar);
  procedure Read(R,A,G: in out Scalar);
  procedure Write(R,A,G: in Scalar);

end Hopf.Fix;
