with Strings, Globals;
use Strings, Globals;

pragma Elaborate_All (Strings,Globals);

package body NSP is

  function Used(I: Indx) return Boolean is
    --- for standard NSPoly
  begin
    return I/=0 or else NN<3;
  end Used;

  function TrueN(I: Indx) return Integer is
    --- for standard NSPoly
  begin
    if I>=0 then return I+NNOffset; end if;
    return I-NNOffset;
  end TrueN;

  --- NS specific Fourier polynomial
  --- basic

  function IsSharp(P: NSPoly) return Boolean is
  begin
    if Not_STrunc then
      for I in Indx loop
        if not IsSharp(P(I)) then return False; end if;
      end loop;
    end if;
    return True;
  end IsSharp;

  function IsZero(P: NSPoly) return Boolean is
  begin
    for I in Indx loop
      if not IsZero(P(I)) then return False; end if;
    end loop;
    return True;
  end IsZero;

  procedure SetZero(P: in out NSPoly) is
  begin
    for I in Indx loop SetZero(P(I)); end loop;
  end SetZero;

  function PolyZero return NSPoly is
    P: NSPoly;
  begin
    for I in Indx loop SetZero(P(I)); end loop;
    return P;
  end PolyZero;

  procedure Copy(P1: in NSPoly; P2: in out NSPoly) is
  begin
    for I in Indx loop Copy(P1(I),P2(I)); end loop;
  end Copy;

  procedure Swap(P1,P2: in out NSPoly) is
  begin
    for I in Indx loop Swap(P1(I),P2(I)); end loop;
  end Swap;

  --- addition and multiplication etc.

  procedure Neg(P: in out NSPoly) is
  begin
    for I in Indx loop Neg(P(I)); end loop;
  end Neg;

  procedure Neg(P1: in NSPoly; P2: in out NSPoly) is
  begin
    for I in Indx loop Neg(P1(I),P2(I)); end loop;
  end Neg;

  function "-"(P: NSPoly) return NSPoly is
    Q: NSPoly;
  begin
    Neg(P,Q);
    return Q;
  end "-";

  procedure Add(P1: in NSPoly; P2: in out NSPoly) is
  begin
    for I in Indx loop Add(P1(I),P2(I)); end loop;
  end Add;

  procedure Sum(P1,P2: in NSPoly; P3: in out NSPoly) is
  begin
    for I in Indx loop Sum(P1(I),P2(I),P3(I)); end loop;
  end Sum;

  function "+"(P1,P2: NSPoly) return NSPoly is
    P3: NSPoly;
  begin
    Sum(P1,P2,P3);
    return P3;
  end "+";

  procedure Sub(P1: in NSPoly; P2: in out NSPoly) is
  begin
    for I in Indx loop Sub(P1(I),P2(I)); end loop;
  end Sub;

  procedure Diff(P1,P2: in NSPoly; P3: in out NSPoly) is
  begin
    for I in Indx loop Diff(P1(I),P2(I),P3(I)); end loop;
  end Diff;

  function "-"(P1,P2: NSPoly) return NSPoly is
    P3: NSPoly;
  begin
    Diff(P1,P2,P3);
    return P3;
  end "-";

  procedure Mult(R: in Flt; P: in out NSPoly; Tmp: in out Scalar) is
  begin
    for I in Indx loop Mult(R,P(I),Tmp); end loop;
  end Mult;

  procedure Prod(R: in Flt; P1: in NSPoly; P2: in out NSPoly) is
  begin
    for I in Indx loop Prod(R,P1(I),P2(I)); end loop;
  end Prod;

  function "*"(R: Flt; P: NSPoly) return NSPoly is
    Q: NSPoly;
  begin
    Prod(R,P,Q);
    return Q;
  end "*";

  procedure AddProd(R: in Flt; P1: in NSPoly; P2: in out NSPoly; Tmp: in out Scalar) is
  begin
    if R/=Zero then
      for I in Indx loop AddProd(R,P1(I),P2(I),Tmp); end loop;
    end if;
  end AddProd;

  procedure Mult(S: in Scalar; P: in out NSPoly) is
  begin
    for I in Indx loop Mult(S,P(I)); end loop;
  end Mult;

  procedure Prod(S: in Scalar; P1: in NSPoly; P2: in out NSPoly) is
  begin
    for I in Indx loop Prod(S,P1(I),P2(I)); end loop;
  end Prod;

  function "*"(S: Scalar; P: NSPoly) return NSPoly is
    Q: NSPoly;
  begin
    Prod(S,P,Q);
    return Q;
  end "*";

  procedure AddProd(S: in Scalar; P1: in NSPoly; P2: in out NSPoly) is
  begin
    if not IsZero(S) then
      for I in Indx loop AddProd(S,P1(I),P2(I)); end loop;
    end if;
  end AddProd;

  procedure Quot(P1: in NSPoly; R: in Flt; P2: in out NSPoly) is
  begin
    for I in Indx loop Quot(P1(I),R,P2(I)); end loop;
  end Quot;

  --  2*cos(1)*cos(n) =  cos(n-1) + cos(n+1)
  --  2*sin(1)*sin(n) =  cos(n-1) - cos(n+1)
  --  2*cos(1)*sin(n) =  sin(n-1) + sin(n+1)
  --  2*sin(1)*cos(n) = -sin(n-1) + sin(n+1)

  procedure Prod(P1,P2: in NSPoly; P3: in out NSPoly) is
    --- assuming   P1=P1(0)+P1(1)Cos(t)+P1(-1)Sin(t) and
    --- NN=1:      P2=P2(0)+P2(1)Cos(t)+P2(-1)Sin(t)
    --- NN>1 odd:  P2=P2(1)Cos(NN*t)+P2(-1)Sin(NN*t)
    --- NN=0:      P2=P2(0)
    --- NN>0 even: P2=P2(2)Cos(NN*t)+P2(-2)Sin(NN*t)
    C0: Scalar renames P1(0);
  begin
    if NNOdd then
      Prod(C0,P2( 1),P3( 1));
      Prod(C0,P2(-1),P3(-1));
      if NN=1 then
        Prod(C0,P2(0),P3(0));
        AddProd(P1( 1),P2(0),P3( 1));
        AddProd(P1(-1),P2(0),P3(-1));
      else
        SetZero(P3(0));
      end if;
    elsif NN=0 then
      Prod(C0,P2(0),P3(0));
    else                    --- NN>0 even
      declare
        C1: Scalar renames P1( 1);
        S1: Scalar renames P1(-1);
        CN: Scalar renames P2( 2);
        SN: Scalar renames P2(-2);
        A:  Scalar renames P3( 2);
        B:  Scalar renames P3(-2);
      begin
        Prod(C1,CN,A);      ---2* A*cos(1)cos(n) = A*cos(n-1) + A*cos(n+1)
        Prod(S1,SN,B);      ---2* B*sin(1)sin(n) = B*cos(n-1) - B*cos(n+1)
        Mult(SHalf,A);
        Mult(SHalf,B);
        Sum( A,B,P3(1));    ---                    sum
        Diff(A,B,P3(3));    ---                                 diff

        Prod(C1,SN,A);      ---2* A*cos(1)sin(n) = A*sin(n+1) + A*sin(n-1)
        Prod(S1,CN,B);      ---2* B*sin(1)cos(n) = B*sin(n+1) - B*sin(n-1)
        Mult(SHalf,A);
        Mult(SHalf,B);
        Sum( A,B,P3(-3));   ---                    sum
        Diff(A,B,P3(-1));   ---                                 diff

        Prod(C0,CN,P3( 2)); --- 1*cos(nx)
        Prod(C0,SN,P3(-2)); --- 1*cos(nx)
        SetZero(P3(0));
      end;
    end if;
  end Prod;

  procedure AddProd(P1,P2: in NSPoly; P3,Tmp: in out NSPoly) is
  begin
    Prod(P1,P2,Tmp);
    Add(Tmp,P3);
  end AddProd;

  --- some functions

  function MaxNorm(C,S: Scalar) return Radius is
    RC: constant Radius := MaxNorm(C);
    RS: constant Radius := MaxNorm(S);
    use Flt_EF;
  begin
    if RS=Zero then return RC; end if;
    if RC=Zero then return RS; end if;
    return Sqrt(RC*RC+RS*RS);
  end MaxNorm;

  function MaxNorm(P: NSPoly) return Radius is
    --- assuming P(0)=0 if NN>2
    R: Flt;
  begin
    if NNOdd then
      if NN=1 then return MaxNorm(P(0))+Rho*MaxNorm(P(1),P(-1)); end if;
      return RhoNN*MaxNorm(P(1),P(-1));
    end if;
    if NN=0 then return MaxNorm(P(0)); end if;
    R := RhoNNMinus*MaxNorm(P(1),P(-1))+RhoNN*MaxNorm(P(2),P(-2))+RhoNNPlus*MaxNorm(P(3),P(-3));
    if NN=2 then return MaxNorm(P(0))+R; end if;
    return R;
  end MaxNorm;

  procedure Norm(P: in NSPoly; S: in out Scalar) is
    --- assuming P(0)=0 if NN>2
    CC,SS,Tmp: Scalar;
    procedure Norm(C,S: in Scalar; N: in out Scalar) is
    begin
      Sqr(C,CC);
      Sqr(S,SS);
      Add(SS,CC);
      Sqrt(CC,N);
    end Norm;
  begin
    if NNOdd then
      Norm(P(1),P(-1),Tmp);
      Prod(SRhoNN,Tmp,S);
      if NN=1 then Norm(P(0),Tmp); Add(Tmp,S); end if;
    elsif NN=0 then
      Norm(P(0),S);
    else
      Norm(P(1),P(-1),Tmp);
      Prod(SRhoNNMinus,Tmp,S);
      Norm(P(2),P(-2),Tmp);
      AddProd(SRhoNN,Tmp,S);
      Norm(P(3),P(-3),Tmp);
      AddProd(SRhoNNPlus,Tmp,S);
      if NN=2 then Norm(P(0),Tmp); Add(Tmp,S); end if;
    end if;
  end Norm;

  function Norm(P: NSPoly) return Scalar is
    S: Scalar;
  begin
    Norm(P,S);
    return S;
  end Norm;

  procedure Coeff(I: in Indx; P: in NSPoly; S: in out Scalar) is
  begin
    if I=0 and then NN>2 then raise Undefined; end if;
    Copy(P(I),S);
  end Coeff;

  --- conversion and i/o

  procedure Assign(R: in Flt; P: in out NSPoly) is
    --- careful, P(0) may not be Used
  begin
    for I in Indx loop SetZero(P(I)); end loop;
    Assign(R,P(0));
  end Assign;

  procedure Put01(F: in File_Type; P: in NSPoly; Decimal: in Boolean := False) is
    --- saving only frequencies 0 and 1
  begin
    for I in -1 .. 1 loop Put(F,P(I),Decimal); end loop;
  end Put01;

  procedure Get01(F: in File_Type; P: in out NSPoly; Decimal: in Boolean := False) is
    Tmp: Scalar;
  begin
    SetZero(Tmp);
    SetZero(P);
    if NN=0 then Get(F,Tmp,Decimal); else Get(F,P(-1),Decimal); end if;
    Get(F,P(0),Decimal);
    if NN=0 then Get(F,Tmp,Decimal); else Get(F,P( 1),Decimal); end if;
  end Get01;

  procedure Center(P: in out NSPOly) is
  begin
    for I in Indx loop Center(P(I)); end loop;
  end Center;

  procedure Show1(N: in String; P: in NSPoly; Hide0: in Boolean := True) is
    P0: Boolean := True;
  begin
    for I in Indx loop
      if not IsZero(P(I)) then
        P0 := False;
        Show1(N & Strng(TrueN(I),4),P(I),True);
      end if;
    end loop;
    if P0 and then (not Hide0) then Show0(N & "0"); end if;
  end Show1;

  --- NS specific Fourier error

  procedure Add(E1: in NSErr; E2: in out NSErr) is
  begin
    for I in NonnegIndx loop E2(I) := E2(I)+E1(I); end loop;
  end Add;

  procedure Sum(E1,E2: in NSErr; E3: out NSErr) is
  begin
    for I in NonnegIndx loop E3(I) := E1(I)+E2(I); end loop;
  end Sum;

  function "+"(E1,E2: NSErr) return NSErr is
    E3: NSErr;
  begin
    for I in NonnegIndx loop E3(I) := E1(I)+E2(I); end loop;
    return E3;
  end "+";

  procedure Mult(E1: in Radius; E2: in out NSErr) is
  begin
    for I in NonnegIndx loop E2(I) := E1*E2(I); end loop;
  end Mult;

  procedure Prod(E1: in Radius; E2: in NSErr; E3: out NSErr) is
  begin
    for I in NonnegIndx loop E3(I) := E1*E2(I); end loop;
  end Prod;

  function "*"(E1: Radius; E2: NSErr) return NSErr is
    E3: NSErr;
  begin
    for I in NonnegIndx loop E3(I) := E1*E2(I); end loop;
    return E3;
  end "*";

  function "*"(E1: NSErr; E2: Radius) return NSErr is
    E3: NSErr;
  begin
    for I in NonnegIndx loop E3(I) := E1(I)*E2; end loop;
    return E3;
  end "*";

  procedure AddProd(E1: in Radius; E2: in NSErr; E3: in out NSErr) is
  begin
    for I in NonnegIndx loop E3(I) := E3(I)+E1*E2(I); end loop;
  end AddProd;

  function "*"(E1,E2: NSErr) return NSErr is
    C0: Flt renames E1(0);
    E3: NSErr;
  begin
    if NNOdd then
      if NN=1 then
        E3(0) := C0*E2(0);
        E3(1) := C0*E2(1)+E1(1)*E2(0);
      else
        E3(0) := Zero;
        E3(1) := C0*E2(1);
      end if;
    elsif NN=0 then
      E3(0) := C0*E2(0);
    else
      E3(0) := Zero;
      E3(2) := C0*E2(2);
      E3(3) := Half*E1(1)*E2(2);
      E3(1) := InvRho2*E3(3);
    end if;
    return E3;
  end "*";

  function MaxNorm(E: NSErr) return Radius is
  begin
    if NNOdd then
      if NN=1 then return E(0)+E(1); end if;
      return E(1);
    end if;
    if NN=2 then return E(0)+E(1)+E(2)+E(3); end if;
    if NN=0 then return E(0); end if;
    return E(1)+E(2)+E(3);
  end MaxNorm;

  function MaxNorms(P: NSPoly) return NSErr is
    --- convert standard NSPoly to NSErr
    E: NSErr;
  begin
    if NNOdd then
      if NN=1 then E(0) := MaxNorm(P(0)); else E(0) := Zero; end if;
      E(1) := RhoNN*MaxNorm(P(1),P(-1));
    elsif NN=0 then
      E(0) := MaxNorm(P(0));
    else                     --- NN>0 even
      if NN=2 then E(0) := MaxNorm(P(0)); else E(0) := Zero; end if;
      E(1) := RhoNNMinus*MaxNorm(P(1),P(-1));
      E(2) := RhoNN*MaxNorm(P(2),P(-2));
      E(3) := RhoNNPlus*MaxNorm(P(3),P(-3));
    end if;
    return E;
  end MaxNorms;

  function MaxNorms1(P: NSPoly) return NSErr is
    --- convert fixed point NSPoly to NSErr
    E: NSErr;
  begin
    E(0) := MaxNorm(P(0));
    if ILast /= 0 then
      E(1) := Rho*MaxNorm(P(1),P(-1));
      if Ilast /= 1 then
        E(2) := Zero;
        E(3) := Zero;
      end if;
    end if;
    return E;
  end MaxNorms1;

  procedure AddBall(R: in Radius; E: in out NSErr) is
  begin
    E(0) := E(0)+R;
    if NN>0 then E(1) := E(1)+R; end if;
  end AddBall;

  procedure BallsAt0(E: in NSErr; P: in out NSPoly) is
    --- used e.g. to compute coefficients
  begin
    if NNOdd then
      if NN=1 then BallAt0(E(0),P(0)); else SetZero(P(0)); end if;
      BallAt0(InvRhoNN*E(1),P(1));
      Copy(P(1),P(-1));
    elsif NN=0 then
      BallAt0(E(0),P(0));
    else                   --- NN>0 even
      if NN=2 then BallAt0(E(0),P(0)); else SetZero(P(0)); end if;
      BallAt0(InvRhoNNMinus*E(1),P(1));
      Copy(P(1),P(-1));
      BallAt0(InvRhoNN*E(2),P(2));
      Copy(P(2),P(-2));
      BallAt0(InvRhoNNPlus*E(3),P(3));
      Copy(P(3),P(-3));
    end if;
  end BallsAt0;

  procedure Coeff(I: in Indx; P: in NSPoly; E: in NSErr; S: in out Scalar) is
    Iabs: constant Integer := abs(I);
  begin
    if STrunc then
      if I=0 and then NN>2 then raise Undefined; end if;
      SetZero(S);
    elsif I=0 then
      if NN>2 then raise Undefined; end if;
      BallAt0(E(0),S);
    elsif Iabs=1 then
      if NNOdd then
        BallAt0(InvRhoNN*E(1),S);
      else
        BallAt0(InvRhoNNMinus*E(1),S);
      end if;
    elsif Iabs=2 then
      BallAt0(InvRhoNN*E(2),S);
    elsif Iabs=3 then
      BallAt0(InvRhoNNPlus*E(3),S);
    end if;
    Add(P(I),S);
  end Coeff;

  procedure Put01(F: in File_Type; E: in NSErr; Decimal: in Boolean := False) is
    --- saving only frequencies 0 and 1
  begin
    for I in 0 .. 1 loop
      if STrunc then Put(F,Zero,Decimal); else Put(F,E(I),Decimal); end if;
    end loop;
  end Put01;

  procedure Get01(F: in File_Type; E: out NSErr; Decimal: in Boolean := False) is
    Tmp: Radius := Zero;
  begin
    E := EZero;
    Get(F,E(0),Decimal);
    if NN=0 then Get(F,Tmp,Decimal); else Get(F,E(1),Decimal); end if;
  end Get01;

  procedure Show1(N: in String; E: in NSErr; Hide0: in Boolean := True) is
    E0: Boolean := True;
  begin
    for I in NonNegIndx loop
      if E(I)/=Zero then
        E0 := False;
        Show1(N & Strng(TrueN(I),4),E(I),True);
      end if;
    end loop;
    if E0 and then (not Hide0) then Show0(N & "0"); end if;
  end Show1;

end NSP;
