with Ada.Text_IO, Logicals, Ints, Rationals, Flts, Scalars;
use Ada.Text_IO, Logicals, Ints, Rationals, Flts, Scalars;

pragma Elaborate_All (Ada.Text_IO,Logicals,Ints,Rationals,Flts,Scalars);

generic

  NN: in Natural;
  Rho: in Radius;

  type Scalar is private;

  pragma Warnings (Off); --- suppress the "... not referenced" warnings

  --- basic
  with function Info(Dummy: Scalar) return Scalar_Info is <>;
  with function IsSharp(S: Scalar) return Boolean is <>;
  with function IsNumbers(S: Scalar) return Boolean is <>;
  with function IsZero(S: Scalar) return Boolean is <>;
  with function "="(S1,S2: Scalar) return Boolean is <>;
  with procedure SetZero(S: in out Scalar) is <>;
  with procedure Copy(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;

  --- sets
  with function Center0(S: Scalar) return Boolean is <>;
  with function Contains0(S: Scalar) return Logical is <>;
  with function Contains(S1,S2: Scalar) return Logical is <>; -- S1 constains S2
  with procedure BallAt0(R: in Flt; S: in out Scalar) is <>;
  with function BallAt0(R: Flt) return Scalar is <>;
  with procedure DiskAt0(R: in Flt; S: in out Scalar) is <>;
  with function DiskAt0(R: Flt) return Scalar is <>;
  with procedure ToErr(S: in out Scalar) is <>;
  with procedure ToErr(S1: in Scalar; S2: in out Scalar) is <>;
  with function ToErr(S: Scalar) return Scalar is <>;
  with procedure Center(S: in out Scalar) is <>;
  with procedure Center(S1: in Scalar; S2: in out Scalar) is <>;
  with function Center(S: Scalar) return Scalar is <>;
  with procedure CenterDisk(S: in out Scalar) is <>;
  with procedure CenterDisk(S1: in Scalar; S2: in out Scalar) is <>;
  with function CenterDisk(S: Scalar) return Scalar is <>;
  with procedure ModCenter(S: in out Scalar) is <>;
  with procedure ModCenter(S1: in Scalar; S2: in out Scalar) is <>;
  with function ModCenter(S: Scalar) return Scalar is <>;
  with procedure ModNumbers(S: in out Scalar) is <>;
  with procedure ModNumbers(S1: in Scalar; S2: in out Scalar) is <>;
  with function ModNumbers(S: Scalar) return Scalar is <>;
  with procedure ErrMult(R: in Radius; S: in out Scalar) is <>;
  with procedure Union(S1: in Scalar; S2: in out Scalar) is <>;
  with function Union(S1,S2: Scalar) return Scalar is <>;
  with procedure Intersection(S1: in Scalar; S2: in out Scalar; Empty: out Logical) is <>;

  --- order
  with function Sign(S: Scalar) return Integer is <>;
  with function Compare(S1,S2: Scalar) return Integer is <>;
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function "<="(S1,S2: Scalar) return Boolean is <>;
  with function ">"(S1,S2: Scalar) return Boolean is <>;
  with function ">="(S1,S2: Scalar) return Boolean is <>;
  with procedure Min(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Min(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Min(S1,S2: Scalar) return Scalar is <>;
  with procedure Max(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Max(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;
  with function Inf(S: Scalar) return Flt is <>; -- inf of real part
  with function Sup(S: Scalar) return Flt is <>; -- sup of real part

  --- arithmetic
  with procedure Neg(S: in out Scalar) is <>;
  with procedure Neg(S1: in Scalar; S2: in out Scalar) is <>;
  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Add(I: in Integer; S: in out Scalar) is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sum(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "+"(S1,S2: Scalar) return Scalar is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Diff(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "-"(S1,S2: Scalar) return Scalar is <>;
  with procedure Mult(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Mult(R: in Flt; S: in out Scalar) is <>;
  with procedure Prod(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with function "*"(R: Flt; S: Scalar) return Scalar is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2,Tmp: in out Scalar) is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Mult(Q: in Rational; S: in out Scalar) is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Prod(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "*"(S1,S2: Scalar) return Scalar is <>;
  with procedure AddProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure AddProd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure SubProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S: in out Scalar) is <>;
  with procedure Quot(S1: in Scalar; R: in Flt; S2: in out Scalar) is <>;
  with function "/"(S: Scalar; R: Flt) return Scalar is <>;
  with procedure Div(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Quot(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "/"(S1,S2: Scalar) return Scalar is <>;
  with procedure Inv(S1: in Scalar; S2: in out Scalar) is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "**"(S: Scalar; I: Integer) return Scalar is <>;

  --- fun
  with function IsReal(S: Scalar) return Boolean is <>;
  with procedure Adjoint(S: in out Scalar) is <>;
  with procedure Adjoint(S1: in Scalar; S2: in out Scalar) is <>;
  with function Adjoint(S: Scalar) return Scalar is <>;
  with procedure Real_Part(S: in out Scalar) is <>;
  with procedure Imag_Part(S: in out Scalar) is <>;
  with procedure Eval0(S: in out Scalar) is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with procedure Norm(S1: in Scalar; S2: in out Scalar) is <>;
  with function Norm(S: Scalar) return Scalar is <>;
  with function MaxNorm(S: Scalar) return Radius is <>;
  with procedure Sqr(S: in out Scalar) is <>;
  with procedure Sqr(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with procedure Sqrt(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;
  with procedure Root(K: in Positive; S1: in Scalar; S2: in out Scalar) is <>;
  with function Root(K: Positive; S: Scalar) return Scalar is <>;
  with procedure Exp(S1: in Scalar; S2: in out Scalar) is <>;
  with function Exp(S: Scalar) return Scalar is <>;
  with procedure ArcCos(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcCos(S: Scalar) return Scalar is <>;
  with procedure ArcSin(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcSin(S: Scalar) return Scalar is <>;
  with procedure Cos(S1: in Scalar; S2: in out Scalar) is <>;
  with function Cos(S: Scalar) return Scalar is <>;
  with procedure Sin(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sin(S: Scalar) return Scalar is <>;
  with procedure Log(S1: in Scalar; S2: in out Scalar) is <>;
  with function Log(S: Scalar) return Scalar is <>;
  with function Pi return Scalar is <>;
  with procedure Simple_Random(S: in out Scalar) is <>;
  with procedure QPower(Q: in Rational; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Roots2(B,C: in Scalar; U1,U2,V: in out Scalar) is <>;
  with procedure Roots3(B,C,D: in Scalar; U0,U1,U2,V: in out Scalar) is <>;

  --- conversion and i/o
  with function Approx(S: Scalar) return Flt is <>;
  with procedure Assign(K: in Integer; S: in out Scalar) is <>;
  with procedure Assign(Q: in Rational; S: in out Scalar) is <>;
  with procedure Assign(R: in Flt; S: in out Scalar) is <>;
  with function Scal(I: Integer) return Scalar is <>;
  with function Scal(Q: Rational) return Scalar is <>;
  with function Scal(R: Flt) return Scalar is <>;
  with procedure Enclose(R1,R2: in Flt; S: in out Scalar) is <>;
  with function Enclose(R1,R2: Flt) return Scalar is <>;
  with procedure Show1(N: in String; S: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Scalar; Decimal: in Boolean := False) is <>;

  --- misc
  with function Get_Precision(R: Scalar) return Positive is <>;
  with procedure Proper_Rounding is <>;
  with procedure Proper_Rounding(Dummy: in Scalar) is <>;
  with procedure Free_Cache(S: in Scalar) is <>;
  -------------------- end_include scalars

  pragma Warnings (On);

package NSP is

  ILast_Even:    constant Natural := Choose(NN=0,0,3);
  ILast_Odd:     constant Natural := 1;
  NNOffset_Even: constant Natural := IMax(0,NN-2);
  NNOffset_Odd:  constant Natural := IMax(0,NN-1);

  NNEven:      constant Boolean := (NN mod 2) = 0;
  NNOdd:       constant Boolean := not NNEven;
  NegNN:       constant Integer := -NN;
  ILast:       constant Natural := Choose(NNEven,ILast_Even,ILast_Odd);
  IFirst:      constant Integer := -ILast;
  NNOffset:    constant Natural := Choose(NNEven,NNOffset_Even,NNOffset_Odd);

  --- NS specific Fourier polynomial

  --  NN ILast NNOffset           frequencies
  --   0   0      0      stat         [0]        true freq
  --   1   1      0      bif       [-1,0,1]      true freq
  --   2   3      0      n=2 [-3,-2,-1,0,1,2,3]  true freq
  --   3   1      2      n=3       [-3,*,3,]
  --   4   3      2      n=4 [-5,-4,-3,*,3,4,5]
  --   5   1      4      n=5       [-5,*,5,]

  subtype Indx is Integer range IFirst .. ILast;
  type NSPoly is array(Indx) of Scalar;

  function Used(I: Indx) return Boolean;
  function TrueN(I: Indx) return Integer;

  --- basic
  function IsSharp(P: NSPoly) return Boolean;         --- polynomial with sharp coeffs
  function IsZero(P: NSPoly) return Boolean;
  procedure SetZero(P: in out NSPoly);
  procedure Copy(P1: in NSPoly; P2: in out NSPoly);
  procedure Swap(P1,P2: in out NSPoly);

  --- addition and multiplication etc.
  procedure Neg(P: in out NSPoly);
  procedure Neg(P1: in NSPoly; P2: in out NSPoly);
  function "-"(P: NSPoly) return NSPoly;
  procedure Add(P1: in NSPoly; P2: in out NSPoly);
  procedure Sum(P1,P2: in NSPoly; P3: in out NSPoly);
  function "+"(P1,P2: NSPoly) return NSPoly;
  procedure Sub(P1: in NSPoly; P2: in out NSPoly);
  procedure Diff(P1,P2: in NSPoly; P3: in out NSPoly);
  function "-"(P1,P2: NSPoly) return NSPoly;
  procedure Mult(R: in Flt; P: in out NSPoly; Tmp: in out Scalar);
  procedure Prod(R: in Flt; P1: in NSPoly; P2: in out NSPoly);
  function "*"(R: Flt; P: NSPoly) return NSPoly;
  procedure AddProd(R: in Flt; P1: in NSPoly; P2: in out NSPoly; Tmp: in out Scalar);
  procedure Mult(S: in Scalar; P: in out NSPoly);
  procedure Prod(S: in Scalar; P1: in NSPoly; P2: in out NSPoly);
  function "*"(S: Scalar; P: NSPoly) return NSPoly;
  procedure AddProd(S: in Scalar; P1: in NSPoly; P2: in out NSPoly);
  procedure Quot(P1: in NSPoly; R: in Flt; P2: in out NSPoly);

  procedure Prod(P1,P2: in NSPoly; P3: in out NSPoly);        --- nonstandard
  procedure AddProd(P1,P2: in NSPoly; P3,Tmp: in out NSPoly); --- nonstandard

  --- some functions
  function MaxNorm(C,S: Scalar) return Radius;
  function MaxNorm(P: NSPoly) return Radius;
  procedure Norm(P: in NSPoly; S: in out Scalar);
  function Norm(P: NSPoly) return Scalar;
  procedure Coeff(I: in Indx; P: in NSPoly; S: in out Scalar);

  --- conversion and i/o
  procedure Assign(R: in Flt; P: in out NSPoly);
  procedure Put01(F: in File_Type; P: in NSPoly; Decimal: in Boolean := False);
  procedure Get01(F: in File_Type; P: in out NSPoly; Decimal: in Boolean := False);
  procedure Center(P: in out NSPOly);
  procedure Show1(N: in String; P: in NSPoly; Hide0: in Boolean := True);

  --- NS specific Fourier error

  subtype NonnegIndx is Integer range 0 .. ILast;
  type NSErr is array(NonnegIndx) of Flt;

  procedure Add(E1: in NSErr; E2: in out NSErr);
  procedure Sum(E1,E2: in NSErr; E3: out NSErr);
  function "+"(E1,E2: NSErr) return NSErr;
  procedure Mult(E1: in Radius; E2: in out NSErr);
  procedure Prod(E1: in Radius; E2: in NSErr; E3: out NSErr);
  function "*"(E1: Radius; E2: NSErr) return NSErr;
  function "*"(E1: NSErr; E2: Radius) return NSErr;
  function "*"(E1,E2: NSErr) return NSErr;
  procedure AddProd(E1: in Radius; E2: in NSErr; E3: in out NSErr);

  function MaxNorm(E: NSErr) return Radius;
  function MaxNorms(P: NSPoly) return NSErr;
  function MaxNorms1(P: NSPoly) return NSErr;
  procedure AddBall(R: in Radius; E: in out NSErr);
  procedure BallsAt0(E: in NSErr; P: in out NSPoly);
  procedure Coeff(I: in Indx; P: in NSPoly; E: in NSErr; S: in out Scalar);
  procedure Put01(F: in File_Type; E: in NSErr; Decimal: in Boolean := False);
  procedure Get01(F: in File_Type; E: out NSErr; Decimal: in Boolean := False);
  procedure Show1(N: in String; E: in NSErr; Hide0: in Boolean := True);

  --- upper bounds
  RhoNN:         constant Radius := NPower(Rho,NN);
  RhoNNMinus:    constant Radius := NPower(Rho,IMax(NN-1,0));
  RhoNNPlus:     constant Radius := NPower(Rho,NN+1);
  InvRho:        constant Radius := One/Rho;
  InvRho2:       constant Radius := InvRho*InvRho;
  InvRhoNN:      constant Radius := NPower(InvRho,NN);
  InvRhoNNMinus: constant Radius := NPower(InvRho,IMax(NN-1,0));
  InvRhoNNPlus:  constant Radius := NPower(InvRho,NN+1);

  SRho:        constant Scalar := Scal(Rho);
  SRhoNN:      constant Scalar := SRho**NN;
  SRhoNNMinus: constant Scalar := SRho**IMax(NN-1,0);
  SRhoNNPlus:  constant Scalar := SRho**(NN+1);

  EZero: constant NSErr := (others => Zero);

private

  function PolyZero return NSPoly;

  SOne:       constant Scalar  := Scal(1);
  SHalf:      constant Scalar  := Scal(Half);
  SInfo:  constant Scalar_Info := Info(SHalf);
  STrunc:     constant Boolean := SInfo.IsNumeric;
  Not_Strunc: constant Boolean := not STrunc;

end NSP;
